

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <map>
#include <iterator>
#include "plink.h"
#include "options.h"

class Pool
{
public:  
  set<Segment*> segs;
};


bool segsOverlap(Segment * s1, Segment * s2)
{
  
  if ( s1->finish < s2->start ) return false;
  else if ( s2->finish < s1->start ) return false;
  else return true;  
}


bool segsMatch(Segment * s1, Segment * s2)
{

  int mismatch = 0;
  
  int start = s1->start > s2->start ? s1->start : s2->start;
  int finish = s1->finish < s2->finish ? s1->finish : s2->finish;
  
  // Assume individual major mode
  
  for (int l=start; l<= finish; l++)
    {
      // Only test homozygotes
      if ( s1->p1->one[l] == s1->p1->two[l] && 
	   s2->p1->one[l] == s2->p1->two[l] )
	if ( s1->p1->one[l] != s2->p1->one[l] )
	  mismatch++;
    }

  if ((double)mismatch/(double)(finish-start+1) > 1-par::fuzzy_homo ) return false;
  else return true;
}



bool spawnNewPool(Pool p1, Segment * s)
{

  set<Segment*>::iterator s2 = p1.segs.begin();
  while ( s2 != p1.segs.end() )
    {
      
      // Already in?
      if ( *s2 == s)
	return false;
      
      // Overlap?
      if ( ! segsOverlap( *s2 , s ) )
	return false;
      
      // Match?
      if (par::homo_summary_allelic_match)
	{
	  if ( ! segsMatch( *s2 , s ) )
	    return false;
	}

      s2++;
    }
  
  return true;
  
}



void mergeSegments(Pool p, vector<Segment*> & possible_segments, vector<Pool> & pools)
{
  // Consider all other segments
  
  vector<Segment*>::iterator s2 = possible_segments.begin();
  int c=1;
  while ( s2 != possible_segments.end() )
    {
      
      // Add new segment, recursively?
      if ( spawnNewPool(p,*s2) )
	{
	  p.segs.insert(*s2);
	  mergeSegments(p,possible_segments,pools);
	}
      s2++;
    }
  
  // Worth adding?
  if ( p.segs.size() > 1 )
    {
      
      // Unqiue?
      bool okay = true;
      for (int e=0; e<pools.size(); e++)
	{
	  bool uniq = false;
	  set<Segment*>::iterator ss = p.segs.begin();
	  while ( ss != p.segs.end() )
	    {
	      if ( pools[e].segs.find( *ss ) == pools[e].segs.end() )
		uniq = true;
	      ss++;
	    }
	  if (!uniq) okay = false;
	}
      
      if (okay)
	pools.push_back(p);
      
    }
}



void Plink::findHomoRuns(Individual * person, ofstream & HOM)
{

  int l=0; 
  int lasthom=0;
  int nmiss = 0;
  int nhet = 0;
  bool run = false;
  int start = 0;
  int end = 0;

  while ( l < nl_all )
    {
      
      // Skip haploid chromosomes / end any existing run
      if ( ( par::chr_sex[locus[l]->chr] && person->sex ) || 
	   par::chr_haploid[locus[l]->chr] ) 
	{
	  if (run) 
	    {
	      end = l-1;
	      run = false;
	    }
	  else
	    {
	      l++;
	      continue; 
	    }
	}
      
      // Outside of a run?
      if (!run)
	{
	  // A new run?
	  if (person->one[l] == person->two[l])
	    {
	      start = lasthom = l;
	      nmiss=0;
	      nhet=0;
	      run=true;
	    }
	}
      else // if already in a run, either end or increase length?
	{

	  if ( locus[l]->chr != locus[start]->chr ) // different chromosome?
	    {
	      end = l-1;
	      run = false;
	    }
	  else if ( l == (nl_all -1) ) // or end of all SNPs?
	    {
	      end = l;
	      run = false;
	    }
	  // found a het?
	  else if ( (!person->one[l]) && person->two[l])
	    {
	      if (nhet==par::homo_run_het)
		{
		  end = lasthom;
		  run = false;
		}
	      else
		nhet++;
	    }
	  else // ...continue run
	    {
	      lasthom=l;
	      if ( person->one[l] && (!person->two[l]) ) 
		nmiss++;
	    }
	  
	}


      // Check run length?
      if (!run)
	{
	  if (par::homo_run_length_kb)
	    {
	      if ( locus[end]->bp - locus[start]->bp >= 
		   par::homo_run_length * 1000 )
		{
		  HOM << setw(par::pp_maxfid) << person->fid << " "
		      << setw(par::pp_maxiid) << person->iid << " "
		      << setw(4) << locus[start]->chr << " "
		      << setw(par::pp_maxsnp) << locus[start]->name << " "
		      << setw(par::pp_maxsnp) << locus[end]->name << " "
		      << setw(12) << locus[start]->bp << " "
		      << setw(12) << locus[end]->bp << " "
		      << setw(10) << (double)(locus[end]->bp - locus[start]->bp)
		    /(double)1000 << " "
		      << setw(10) << end - start +1 << " "
		      << setw(4) << nhet << " "
		      << setw(4) << nmiss << "\n";
		  
		  Segment s;
		  s.p1 = s.p2 = person;
		  s.start = start;
		  s.finish = end;
		  segment.push_back(s);
		}
	    }
	  else 
	    {
	      if ( end - start +1 >= par::homo_run_length  )
		{
		  HOM << setw(par::pp_maxfid) << person->fid << " "
		      << setw(par::pp_maxiid) << person->iid << " "
		      << setw(4) << locus[start]->chr << " "
		      << setw(par::pp_maxsnp) << locus[start]->name << " "
		      << setw(par::pp_maxsnp) << locus[end]->name << " "
		      << setw(12) <<locus[start]->bp << " "
		      << setw(12) <<locus[end]->bp << " "
		      << setw(10) << (double)(locus[end]->bp - locus[start]->bp)/(double)1000 << " "
		      << setw(10) <<end - start + 1 << " "
		      << setw(4) << nhet << " "
		      << setw(4) << nmiss << "\n";
		  
		  Segment s;
		  s.p1 = s.p2 = person;
		  s.start = start;
		  s.finish = end;
		  segment.push_back(s);
		}

	    }
	  
	  
	  //////////////////
	  // Clear counters

	  start = end = nmiss = 0;
	  
	}
      
      ///////////////
      // Next locus

      l++;
    }

}




void Plink::summariseHomoRuns()
{

  string f = par::output_file_name + ".hom.summary";
  ofstream HOM(f.c_str(),ios::out);
  HOM.precision(6);
  printLOG("Writing homozygosity-run summary to [ "+f+" ] \n");
  

  if (par::fuzzy_homo > 1e-6) 
    printLOG("Requiring allelic match, threshold "+dbl2str(par::fuzzy_homo)+" identity\n");
  else
    {
      printLOG("Not requiring allelic match (overlapping only)\n");
      par::fuzzy_homo = 0;
    }

  
  HOM << setw(par::pp_maxfid) << "FID" << " "
      << setw(par::pp_maxiid) << "IID" << " "
      << setw(4) << "CHR" << " " 
      << setw(par::pp_maxsnp) << "SNP1" << " " 
      << setw(par::pp_maxsnp) << "SNP2" << " "
      << setw(14) << "BP1" << " " 
      << setw(14) << "BP2" << " "
      << setw(14) << "KB" << "\n";

  // Make groups of segments
  // Group by maximally over-lapping region
  
  // Both irrespective of alleles
  // And considering alleles
  
  // i.e. all segments in a group must have at least one position where they overlap
  // AND (optinally) all overlapping alleles must be the same
  
  // Each segment can be in more than one group
  
  //   ----                  -----
  // ----                 ----------
  //   -----------------------
  //   |                     |


  // Collections of segments
  vector<Pool> pools;


  // Take each segment

  vector<Segment>::iterator s = segment.begin();
  int c = 1;
  while ( s != segment.end() )
    {      
      
      Pool p;
      p.segs.insert(&(*s));
      
      // Construct list of possible segments
      vector<Segment*> possible_segments;
      vector<Segment>::iterator s2 = segment.begin();
      while (  s2 != segment.end() )
	{
	  // Add new segment, recursively?
	  if ( spawnNewPool(p,&(*s2)) )
	    possible_segments.push_back(&(*s2));
	  s2++;
	}
      
      // Begin recursive procedure
      mergeSegments(p,possible_segments,pools);
      s++;
      c++;
    }

   for (int i=0; i<pools.size(); i++)
    {
      set<Segment*>::iterator ss = pools[i].segs.begin();
      
      int min = 0;
      int max = nl_all;
      
      while ( ss != pools[i].segs.end() )
	{
	  
	  min = (*ss)->start > min ? (*ss)->start : min;
	  max = (*ss)->finish < max ? (*ss)->finish : max;

	  HOM << setw(par::pp_maxfid) << (*ss)->p1->fid << " "
	      << setw(par::pp_maxiid) << (*ss)->p1->iid << " "
	      << setw(4) << locus[(*ss)->start]->chr << " "
	      << setw(par::pp_maxsnp) << locus[(*ss)->start]->name << " "
	      << setw(par::pp_maxsnp) << locus[(*ss)->finish]->name << " "
	      << setw(14) << locus[(*ss)->start]->bp << " "
	      << setw(14) << locus[(*ss)->finish]->bp << " "
	      << setw(14) << (double)(locus[(*ss)->finish]->bp - locus[(*ss)->start]->bp ) / 1000.0 
	      << "\n";
	  
	  ss++;
	}
      HOM << setw(par::pp_maxfid) << "NA" << " "
	  << setw(par::pp_maxiid) << "NA" << " "
	  << setw(4) << "NA" << " "
	  << setw(par::pp_maxsnp) << locus[min]->name << " "
	  << setw(par::pp_maxsnp) << locus[max]->name << " "
	  << setw(14) << locus[min]->bp << " "
	  << setw(14) << locus[max]->bp << " "
	  << setw(14) << (double)(locus[max]->bp - locus[min]->bp ) / 1000.0 ;
      HOM << "\n\n";
    }

   printLOG("Found "+int2str(pools.size())+" possibly overlapping sets of shared homozygous segments\n");

   HOM.close();  
}

